
// PerLandMap.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, January 2008

/* The class for drawing the land map

   The series of image operations applied here are:

       rotate --> crop --> warp --> scale --> translate

   The rotate code comes from RotLandMap; the other operations are new.
   The land image is warped to give it a perspective projection. 
   The warping utilizes JAI, available from:
      https://jai.dev.java.net/
*/

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.io.*;
import javax.imageio.ImageIO;

import javax.media.jai.*;
import java.awt.image.renderable.ParameterBlock;



public class PerLandMap
{
  // background colour of the land 
  private static final Color LAND_BG_COLOUR =  new Color(115,140,115);  // gray green

  // crop dimensions
  private static final int CROP_WIDTH = 200;
  private static final int CROP_HEIGHT = 300;

  // warp inset along x-axis of top edge
  private static final int INSET = 50;   // 2*INSET must be < CROP_WIDTH

  // scale factors
  private static final float SCALE_WIDTH = 4.0f;
  private static final float SCALE_HEIGHT = 1.3f;


  // panel dimensions
  private int pWidth, pHeight; 

  // land image info
  private BufferedImage landIm;
  private int landWidth, landHeight;
  private WarpPerspective warpOp;   // for warping the land image




  public PerLandMap(String imFnm)
  {
    // calculate panel dimensions
    pWidth = (int)((CROP_WIDTH - 2*INSET) * SCALE_WIDTH);
    pHeight = (int)(CROP_HEIGHT * SCALE_HEIGHT);

    landIm = loadImage(imFnm);   // load land image
    landWidth = landIm.getWidth();
    landHeight = landIm.getHeight();

    createWarpOp();
  } // end of PerLandMap()



  private BufferedImage loadImage(String fn)
  // load the image stored in fn in images/
  {
    BufferedImage im = null;
    try {
      im = ImageIO.read( getClass().getResource("images/"+fn));
    }
    catch(IOException e) {
      System.out.println("Could not load images/" + fn);
      System.exit(0);
    }
    return im;
  }  // end of loadImage()



  private void createWarpOp()
  {
    // create the perspective transformation for use later
    float w = CROP_WIDTH;
    float h = CROP_HEIGHT;
    PerspectiveTransform perspTrans = 
                  PerspectiveTransform.getQuadToQuad(
                     0,0,     w,0,        w,h, 0,h, 
                     INSET,0, w-INSET,0,  w,h, 0,h  );
      // warp the top edge of the image inwards along the x-axis
    warpOp = null;
    try {
      warpOp = new WarpPerspective( perspTrans.createInverse() );   // invert the transform
    }
    catch(Exception e)
    {  System.out.println(e);  
       System.exit(1);
    }
  }  // end of createWarpOp()



  public Dimension getPanelDim()
  // return the panel dimensions
  {  return new Dimension(pWidth, pHeight);  }


  public Point2D.Double getStartPos()
  // the player starts in the center of the land 
  {  return new Point2D.Double(landWidth/2, landHeight/2);  }



  // ---------------- drawing the scene --------------------------


  public void draw(Graphics2D g2d, double xPlayer, double yPlayer, int angle)
  // draw a rotated, cropped, warped, scaled, translated land image (!)
  {
    // smooth any output
    g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
                         RenderingHints.VALUE_INTERPOLATION_BILINEAR);

    // rotation around image center
    AffineTransform at = AffineTransform.getRotateInstance(
                                        Math.toRadians(-angle),   // -ve rotation
                                        landWidth/2, landHeight/2);

    BufferedImage rotIm = rotateImage(landIm, at, angle);
    Point rotCoord = calcRotatedPosn(at, xPlayer, yPlayer, angle);

    BufferedImage cropIm = cropImage(rotIm, rotCoord);      // crop
    BufferedImage warpIm = warpImage(cropIm);     // warp

    // draw a scaled and translated version of the land
    drawScaleTrans(g2d, warpIm);
  } // end of draw()



  private BufferedImage rotateImage(BufferedImage im, AffineTransform at, int angle)
  // rotate the land image
  {
    BufferedImage rotIm = im;
    if (angle != 0) {  // only do this if the image needs rotating
      AffineTransformOp rotateOp = 
          new AffineTransformOp(at, AffineTransformOp.TYPE_BILINEAR);
      rotIm = rotateOp.filter(im, null);
    }
    return rotIm;
  }  // end of rotateImage()


  private Point calcRotatedPosn(AffineTransform at, double xPlayer, double yPlayer, int angle)
  // calculate the position of the player on the rotated land
  {
    Point rotCoord = new Point();

    if (angle != 0) {  // only do this if the position needs rotating
      // rotate the player position
      Point2D.Double rotPt = 
          (Point2D.Double) at.transform( 
                  (Point2D)( new Point2D.Double(xPlayer, yPlayer)), null);
      rotCoord.x = (int) rotPt.x;
      rotCoord.y = (int) rotPt.y;
    }
    else {
      rotCoord.x = (int) xPlayer;
      rotCoord.y = (int) yPlayer;
    }
    // System.out.println("calcRotatedPosn (x,y): (" + rotCoord.x + ", " + rotCoord.y + ")");

    return rotCoord;
  }  // end of calcRotatedPosn()


  private BufferedImage cropImage(BufferedImage im, Point rotCoord)
  /* First clip the image to be close to the required crop dimensions, 
     then 'fill in' the clipped image edges so it has the required crop 
     dimensions. */
  {
    // top left coord of clipped image
    int xIm = rotCoord.x - CROP_WIDTH/2;
    int yIm = rotCoord.y - CROP_HEIGHT;
    BufferedImage clipIm = clipImage(im, xIm, yIm);

    return fillEdges(clipIm, xIm, yIm);
  }  // end of cropImage()



  private BufferedImage clipImage(BufferedImage im, int xIm, int yIm)
  /* Extract the part of the land image that falls within the
     crop boundaries. This may be CROP_WIDTH x CROP_HEIGHT in size, 
     or perhaps smaller. */
  {
    // calculate the top-left coord of the subimage
    int xClip = (xIm < 0) ? 0 : xIm;      // tests to prevent -ve values
    int yClip = (yIm < 0) ? 0 : yIm;

    // get the width and height of the subimage
    int widthClip, heightClip;
    if (xClip + CROP_WIDTH > im.getWidth())
      widthClip = im.getWidth() - xClip;    // subimage width cannot go beyound image width 
    else
      widthClip = CROP_WIDTH;

    if (yClip + CROP_HEIGHT > im.getHeight())
      heightClip = im.getHeight() - yClip;    // subimage height cannot go beyound image height
    else
      heightClip = CROP_HEIGHT;

    return im.getSubimage(xClip, yClip, widthClip, heightClip);
  }  // end of clipImage



  private BufferedImage fillEdges(BufferedImage clipIm, int xIm, int yIm)
  /* Draw the clip image into a new BufferedImage of the required crop
     dimensions, filling in any blank parts with the background colour, and 
     positioning the image correctly. */
  {
    // store clipIm into a CROP_WIDTH x CROP_HEIGHT size image
    BufferedImage cropped = new BufferedImage(CROP_WIDTH, CROP_HEIGHT,
                                                BufferedImage.TYPE_INT_ARGB);
	Graphics g = cropped.getGraphics();

    // fill the image with the background colour
    g.setColor( LAND_BG_COLOUR );
    g.fillRect(0, 0, CROP_WIDTH, CROP_HEIGHT);

    /* Position the clipped image to deal with the chance that (xIm, yIm) are
       negative. In that case, the clip should be drawn moved to the right 
       and down to include empty background space at the top left. */
    int xCrop = (xIm < 0) ? -xIm : 0;
    int yCrop = (yIm < 0) ? -yIm : 0;
	g.drawImage(clipIm, xCrop, yCrop, null);   // draw clip

    return cropped;
  }  // end of fillEdges()



  private BufferedImage warpImage(BufferedImage im)
  // warp the top of the image inwards by xInset on the top-left and top-right corners
  {
    ParameterBlock pb = new ParameterBlock();
    pb.addSource( PlanarImage.wrapRenderedImage(im) );    // BufferedImage --> PlanarImage
    pb.add( warpOp ); 
    pb.add( new InterpolationBilinear() );   // smooth
    PlanarImage warpedImage = JAI.create("warp", pb);

    BufferedImage warpIm = null;
    try {
      warpIm = warpedImage.getAsBufferedImage();     // PlanarImage --> BufferedImage
    }
    catch (Exception e) 
    { System.out.println(e); }

    return warpIm;
  }  // end of warpImage()



  private void drawScaleTrans(Graphics2D g2d, BufferedImage im)
  // draw a scaled and translated version of the image
  {
    int newW = (int)(CROP_WIDTH * SCALE_WIDTH);
    int newH = (int)(CROP_HEIGHT * SCALE_HEIGHT);

    int transX = (int)(-INSET * SCALE_WIDTH);  
       /* draw image starting off to the left of the panel, 
          so slanted sides are invisible */

    g2d.drawImage(im, transX, 0, newW, newH, null);
  }  // end of drawScaleTrans()

}  // end of PerLandMap class
