
// RotLandMap.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, September 2008

/* The land moves and rotates while the player stays in the center of
   the window, always facing up.
*/

import java.awt.*;
import java.io.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.imageio.ImageIO;


public class RotLandMap
{
  // panel dimensions
  private int pWidth, pHeight; 

  // land image info
  private BufferedImage landIm;
  private int landWidth, landHeight;



  public RotLandMap(String imFnm, int pW, int pH)
  {
    pWidth = pW;
    pHeight = pH;

    landIm = loadImage(imFnm);   // load land image
    landWidth = landIm.getWidth();
    landHeight = landIm.getHeight();
  } // end of RotLandMap()



  private BufferedImage loadImage(String fn)
  // load the image stored in fn in images/
  {
    BufferedImage im = null;
    try {
      im = ImageIO.read( getClass().getResource("images/"+fn));
    }
    catch(IOException e) {
      System.out.println("Could not load images/" + fn);
      System.exit(0);
    }
    return im;
  }  // end of loadImage()



  public Point2D.Double getStartPos()
  // the player starts in the center of the land 
  {  return new Point2D.Double(landWidth/2, landHeight/2);  }




  public void draw(Graphics g, double xPlayer, double yPlayer, int angle)
  // rotate and translate the land image and draw it
  {
    // rotation around image center
    AffineTransform at = AffineTransform.getRotateInstance(
                                        Math.toRadians(-angle),   // -ve rotation
                                        landWidth/2, landHeight/2);

    BufferedImage rotIm = rotateImage(landIm, at, angle);
    Point rotCoord = calcRotatedPosn(at, xPlayer, yPlayer, angle);

    /* (xIm, yIm) is the top-left drawing position on 
       the land image in the JPanel such that 
       (rotCoord.x, rotCoord.y) is at the center of the JPanel */
    int xIm = pWidth/2 - rotCoord.x;
    int yIm = pHeight/2 - rotCoord.y;
    // System.out.println("rotateImage (x,y): (" + xIm + ", " + yIm + ")");

    g.drawImage(rotIm, xIm, yIm, null);  // draw the rotated land, translated to (xIm, yIm)
  }  // end of draw()



  private BufferedImage rotateImage(BufferedImage im, AffineTransform at, int angle)
  // rotate the land image
  {
    BufferedImage rotIm = im;
    if (angle != 0) {  // only do this if the image needs rotating
      AffineTransformOp rotateOp = 
          new AffineTransformOp(at, AffineTransformOp.TYPE_BILINEAR);
      rotIm = rotateOp.filter(im, null);
    }
    return rotIm;
  }  // end of rotateImage()



  private Point calcRotatedPosn(AffineTransform at, double xPlayer, double yPlayer, int angle)
  // calculate the position of the player on the rotated land
  {
    Point rotCoord = new Point();

    if (angle != 0) {  // only do this if the position needs rotating
      // rotate the player position
      Point2D.Double rotPt = 
          (Point2D.Double) at.transform( 
                (Point2D)( new Point2D.Double(xPlayer, yPlayer)), null);
      rotCoord.x = (int) rotPt.x;
      rotCoord.y = (int) rotPt.y;
    }
    else {
      rotCoord.x = (int) xPlayer;
      rotCoord.y = (int) yPlayer;
    }
    // System.out.println("calcRotatedPosn (x,y): (" + rotCoord.x + ", " + rotCoord.y + ")");

    return rotCoord;
  }  // end of calcRotatedPosn()


}  // end of RotLandMap class
