
// TourLandPanel.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, September 2008

/* Move the player over a land image. The player can move
   forwards, backwards, and rotate, but cannot move off the land image.

   The land moves whle the player stays in the center of the window.
   The player image rotates when the player turns.

   Version 1 of a TourLand Panel.
*/

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.io.*;
import javax.imageio.ImageIO;
// import java.text.DecimalFormat;


public class TourLandPanel extends JPanel
{
  // panel dimensions
  private static final int PWIDTH = 500; 
  private static final int PHEIGHT = 400; 

  private static final String LAND_IM = "londonMap.gif";
  private static final String PLAYER_IM = "player.gif";
                                    // racing car, viewed from above

  // background colour
  private static final Color LAND_BG_COLOUR =  new Color(115,140,115);  // gray green

  // direction constants
  private static final int FWD = 0;
  private static final int BACK = 1;
  private static final int TURN_ANGLE = 10;  // default turning angle (in degrees)
  private static final double STEP = 4.0;  // distance to move forward


  // land image info
  private BufferedImage landIm;
  private int landWidth, landHeight;

  // player info
  private BufferedImage playerIm;
  private int xDrawPos, yDrawPos;  // drawing coords for player (always in the center)

  private int xPlayer, yPlayer;  // position of player on the land
  private int angle;       // angle to the vertical; positive is clockwise

  // private DecimalFormat df = new DecimalFormat("0.##");  // 2 dp


  public TourLandPanel()
  {
    setPreferredSize( new Dimension(PWIDTH, PHEIGHT));
    setBackground( LAND_BG_COLOUR ); 

    initImages();

    setFocusable(true);
    requestFocus();    // the JPanel now has focus, so receives key events
	addKeyListener( new KeyAdapter() {
      public void keyPressed(KeyEvent e)
      { processKey(e);  }
    });
  } // end of TourLandPanel()



  private void initImages()
  // load land and player images, and calculate their positions/sizes
  {
    landIm = loadImage(LAND_IM);   // load land image
    landWidth = landIm.getWidth();
    landHeight = landIm.getHeight();

    playerIm = loadImage(PLAYER_IM);  // load player image (the car)

    // player's drawing position in JPanel (fixed)
    //    - at the center of the window
    xDrawPos = PWIDTH/2 - playerIm.getWidth()/2;
    yDrawPos = PHEIGHT/2 - playerIm.getHeight()/2;

    // the player starts in the center of the land 
    xPlayer = landWidth/2;
    yPlayer = landHeight/2;
    angle = 0;   // player facing north
  }  // end of initImages()



  private BufferedImage loadImage(String fn)
  // load the image stored in fn in images/
  {
    BufferedImage im = null;
    try {
      im = ImageIO.read( getClass().getResource("images/"+fn));
    }
    catch(IOException e) {
      System.out.println("Could not load images/" + fn);
      System.exit(0);
    }
    return im;
  }  // end of loadImage()


  // --------------------- processing input -----------------------------


  private void processKey(KeyEvent e)
  // handles termination and game-play keys
  {
    int keyCode = e.getKeyCode();

    // termination keys
	// listen for esc, q, end, ctrl-c on the canvas
    if ((keyCode == KeyEvent.VK_ESCAPE) || (keyCode == KeyEvent.VK_Q) ||
        (keyCode == KeyEvent.VK_END) ||
        ((keyCode == KeyEvent.VK_C) && e.isControlDown()) )
      System.exit(0);
    
    // game-play keys
    if (keyCode == KeyEvent.VK_LEFT)
      turn(-TURN_ANGLE);   // counter clockwise
    else if (keyCode == KeyEvent.VK_RIGHT)
      turn(TURN_ANGLE);   // clockwise
    else if (keyCode == KeyEvent.VK_UP)
      move(FWD);
    else if (keyCode == KeyEvent.VK_DOWN)
      move(BACK);
  }  // end of processKey()


  private void move(int dir)
  /* Move the player in the dir direction.
     Calculate the x- and y- axis components of the step, and add
     it to the current position in xPlayer, yPlayer.
  */
  {
    double rad = Math.toRadians(angle);
    double xStep = STEP * Math.sin(rad);
    double yStep = STEP * Math.cos(rad);

    if (dir == FWD) {   // forwards
      xPlayer += xStep;
      yPlayer -= yStep;
    }
    else if (dir == BACK) {  // backwards
      xPlayer -= xStep;
      yPlayer += yStep;
    }
    else
      System.out.println("Unknown direction: " + dir);

    // keep player on the land
    if (xPlayer < 0)
      xPlayer = 0;
    else if (xPlayer > landWidth-1)
      xPlayer = landWidth-1;

    if (yPlayer < 0)
      yPlayer = 0;
    else if (yPlayer > landHeight-1)
      yPlayer = landHeight-1;

    repaint();
  }  // end of move()


  private void turn(int ang)
  // rotate the player by ang degrees
  {
    angle = (angle + ang) % 360;
    repaint();
  } // end of turn()


  // ---------------- drawing the scene --------------------------


  public void paintComponent(Graphics g)
  // draw the land then the (rotated) player
  {
    super.paintComponent(g);

    /* (xLand,yland) is the top-left drawing position on 
       the land image in the JPanel such that 
       (xPlayer, yPlayer) is at the center of the JPanel */
    int xLand = (int)(PWIDTH/2 - xPlayer);
    int yLand = (int)(PHEIGHT/2 - yPlayer);
    g.drawImage(landIm, xLand, yLand, null);    // draw the land

    // draw the (rotated) player
    BufferedImage rotIm = rotatePlayer(playerIm);
    g.drawImage(rotIm, xDrawPos, yDrawPos, null);
  } // end of paintComponent()



  private BufferedImage rotatePlayer(BufferedImage im)
  // rotate the player image
  {
    BufferedImage rotIm = im;
    if (angle != 0) {  // only do this if the image needs rotating
      // rotate by angle degrees around the center of the image
      AffineTransform at = AffineTransform.getRotateInstance(
                                        Math.toRadians(angle),
                                        playerIm.getWidth()/2, 
                                        playerIm.getHeight()/2);
      AffineTransformOp rotateOp = 
          new AffineTransformOp(at, AffineTransformOp.TYPE_BILINEAR);
      rotIm = rotateOp.filter(im, null);
    }
    return rotIm;
  }  // end of rotatePlayer()

}  // end of TourLandPanel class
