
// TourPerLandPanel.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, September 2008

/* Version 3 of a JPanel for TourLand

   Move the player's car over a fake 3D landscape. The player can move
   forwards, backwards, and rotate, but cannot move off the land image.

   The land image is warped to give it a perspective projection. This image
   moves and rotates while the player image stays on the bottom edge of the
   window.
*/

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.io.*;
import javax.imageio.ImageIO;



public class TourPerLandPanel extends JPanel
{
  // image used
  private static final String LAND_IM = "londonMap2.gif";
      /* A version of londonMap.gif where the map image is 
         fully contained within the rotation radius of the file.
      */
  private static final String PLAYER_IM = "smallRacer.png"; 
                                  // the back of a racing car

  // land limits (determined for londonMap2.gif)
  private static final int MIN_X = 59; 
  private static final int MAX_X = 610; 
  private static final int MIN_Y = 157; 
  private static final int MAX_Y = 518; 

  // direction constants
  private static final int FWD = 0;
  private static final int BACK = 1;
  private static final int TURN_ANGLE = 10;  // default turning angle (in degrees)
  private static final double STEP = 4.0;  // distance to move forward


  private PerLandMap landMap;    // land info

  // player info
  private BufferedImage playerIm;
  private int xDrawPos, yDrawPos;  // drawing coords for player (fixed)

  private double xPlayer, yPlayer;  // position of player on the land
  private int angle;       // angle to the vertical; positive is clockwise



  public TourPerLandPanel()
  {
    initImages();

    setFocusable(true);
    requestFocus();    // the JPanel now has focus, so receives key events
	addKeyListener( new KeyAdapter() {
      public void keyPressed(KeyEvent e)
      { processKey(e);  }
    });
  } // end of TourPerLandPanel()



  private void initImages()
  // initialize the land and player images, and the player's coordinates
  {
    landMap = new PerLandMap(LAND_IM);
    Dimension panelDim = landMap.getPanelDim();
    setPreferredSize(panelDim);

    playerIm = loadImage(PLAYER_IM);  // load player image

    // calculate fixed drawing position for player 
    //    - resting on the bottom edge, in the middle
    xDrawPos = panelDim.width/2 - playerIm.getWidth()/2;
    yDrawPos = panelDim.height - playerIm.getHeight();
        // changed from TourRotLandPanel

    // the player starts in the center of the land 
    Point2D.Double startPt = landMap.getStartPos();
    xPlayer = startPt.x;
    yPlayer = startPt.y;

    angle = 0;   // player facing north
  }  // end of initImages()


  private BufferedImage loadImage(String fn)
  // load the image stored in fn in images/
  {
    BufferedImage im = null;
    try {
      im = ImageIO.read( getClass().getResource("images/"+fn));
    }
    catch(IOException e) {
      System.out.println("Could not load images/" + fn);
      System.exit(0);
    }
    return im;
  }  // end of loadImage()



  // --------------------- processing input -----------------------------


  private void processKey(KeyEvent e)
  // handles termination and game-play keys
  {
    int keyCode = e.getKeyCode();

    // termination keys
	// listen for esc, q, end, ctrl-c on the canvas
    if ((keyCode == KeyEvent.VK_ESCAPE) || (keyCode == KeyEvent.VK_Q) ||
        (keyCode == KeyEvent.VK_END) ||
        ((keyCode == KeyEvent.VK_C) && e.isControlDown()) )
      System.exit(0);
    
    // game-play keys
    if (keyCode == KeyEvent.VK_LEFT)
      turn(-TURN_ANGLE);   // counter clockwise
    else if (keyCode == KeyEvent.VK_RIGHT)
      turn(TURN_ANGLE);   // clockwise
    else if (keyCode == KeyEvent.VK_UP)
      move(FWD);
    else if (keyCode == KeyEvent.VK_DOWN)
      move(BACK);
  }  // end of processKey()


  private void move(int dir)
  /* Move the player in the dir direction.
     Calculate the x- and y- axis components of the step, and add
     it to the current position in xPlayer, yPlayer.
  */
  {
    double rad = Math.toRadians(angle);
    double xStep = STEP * Math.sin(rad);
    double yStep = STEP * Math.cos(rad);

    if (dir == FWD) {   // forwards
      xPlayer += xStep;
      yPlayer -= yStep;
    }
    else if (dir == BACK) {  // backwards
      xPlayer -= xStep;
      yPlayer += yStep;
    }
    else
      System.out.println("Unknown direction: " + dir);

    // keep player on the land
    if (xPlayer < MIN_X)
      xPlayer = MIN_X;
    else if (xPlayer > MAX_X)
      xPlayer = MAX_X;

    if (yPlayer < MIN_Y)
      yPlayer = MIN_Y;
    else if (yPlayer > MAX_Y)
      yPlayer = MAX_Y;

    repaint();
  }  // end of move()


  private void turn(int ang)
  // rotate the player by ang degrees
  {
    angle = (angle + ang) % 360;
    repaint();
  } // end of turn()



  // ---------------- drawing the scene --------------------------


  public void paintComponent(Graphics g)
  // draw land then player
  {
    super.paintComponent(g);
    Graphics2D g2d = (Graphics2D)g;

    landMap.draw(g2d, xPlayer, yPlayer, angle);     // draw warped land image

    g2d.drawImage(playerIm, xDrawPos, yDrawPos, null);  // draw the player
  }  // end of paintComponent()


}  // end of TourPerLandPanel class
