
// TourRotLandPanel.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, September 2008

/* Move the player over a land image. The player can move
   forwards, backwards, and rotate, but cannot move off the land image.

   The land moves and rotates while the player stays in the center of
   the window, always facing up.

   Version 2 of the TourLand Panel.
*/

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.imageio.ImageIO;
import java.text.DecimalFormat;


public class TourRotLandPanel extends JPanel
{
  // panel dimensions
  private static final int PWIDTH = 500; 
  private static final int PHEIGHT = 400; 

  private static final String LAND_IM = "londonMap2.gif";
      /* A version of londonMap.gif where the map image is 
         fully contained within the rotation radius of the file.
      */
  private static final String PLAYER_IM = "player.gif";

  // land limits (determined for londonMap2.gif)
  private static final int MIN_X = 59; 
  private static final int MAX_X = 610; 
  private static final int MIN_Y = 157; 
  private static final int MAX_Y = 518; 

  // background colour
  private static final Color LAND_BG_COLOUR =  new Color(115,140,115);  // gray green

  // direction constants
  private static final int FWD = 0;
  private static final int BACK = 1;
  private static final int TURN_ANGLE = 10;  // default turning angle (in degrees)
  private static final double STEP = 4.0;  // distance to move forward


  private RotLandMap landMap;    // land info

  // player info
  private BufferedImage playerIm;
  private int xDrawPos, yDrawPos;  // drawing coords for player (always in the center)

  private double xPlayer, yPlayer;  // position of player on the land
  private int angle;       // angle to the vertical; positive is clockwise

  private DecimalFormat df = new DecimalFormat("0.##");  // 2 dp


  public TourRotLandPanel()
  {
    setPreferredSize( new Dimension(PWIDTH, PHEIGHT));
    setBackground( LAND_BG_COLOUR ); 

    initImages();

    setFocusable(true);
    requestFocus();    // the JPanel now has focus, so receives key events
	addKeyListener( new KeyAdapter() {
      public void keyPressed(KeyEvent e)
      { processKey(e);  }
    });
  } // end of TourRotLandPanel()



  private void initImages()
  // load land and player images, and set the player position
  {
    landMap = new RotLandMap(LAND_IM, PWIDTH, PHEIGHT);   // load land image

    playerIm = loadImage(PLAYER_IM);  // load player image

    // player's drawing position in JPanel (fixed)
    //    - at the center of the window
    xDrawPos = PWIDTH/2 - playerIm.getWidth()/2;
    yDrawPos = PHEIGHT/2 - playerIm.getHeight()/2;

    // the player starts in the center of the land 
    Point2D.Double startPt = landMap.getStartPos();
    xPlayer = startPt.x;
    yPlayer = startPt.y;

    angle = 0;   // player facing north
  }  // end of initImages()



  private BufferedImage loadImage(String fn)
  // load the image stored in fn in images/
  {
    BufferedImage im = null;
    try {
      im = ImageIO.read( getClass().getResource("images/"+fn));
    }
    catch(IOException e) {
      System.out.println("Could not load images/" + fn);
      System.exit(0);
    }
    return im;
  }  // end of loadImage()


  // --------------------- processing input -----------------------------


  private void processKey(KeyEvent e)
  // handles termination and game-play keys
  {
    int keyCode = e.getKeyCode();

    // termination keys
	// listen for esc, q, end, ctrl-c on the canvas
    if ((keyCode == KeyEvent.VK_ESCAPE) || (keyCode == KeyEvent.VK_Q) ||
        (keyCode == KeyEvent.VK_END) ||
        ((keyCode == KeyEvent.VK_C) && e.isControlDown()) )
      System.exit(0);
    
    // game-play keys
    if (keyCode == KeyEvent.VK_LEFT)
      turn(-TURN_ANGLE);   // counter clockwise
    else if (keyCode == KeyEvent.VK_RIGHT)
      turn(TURN_ANGLE);   // clockwise
    else if (keyCode == KeyEvent.VK_UP)
      move(FWD);
    else if (keyCode == KeyEvent.VK_DOWN)
      move(BACK);
  }  // end of processKey()


  private void move(int dir)
  /* Move the player in the dir direction.
     Calculate the x- and y- axis components of the step, and add
     it to the current position in xPlayer, yPlayer.
  */
  {
    double rad = Math.toRadians(angle);
    double xStep = STEP * Math.sin(rad);
    double yStep = STEP * Math.cos(rad);

    if (dir == FWD) {   // forwards
      xPlayer += xStep;
      yPlayer -= yStep;
    }
    else if (dir == BACK) {  // backwards
      xPlayer -= xStep;
      yPlayer += yStep;
    }
    else
      System.out.println("Unknown direction: " + dir);

    // keep player on the land (changed from TourLandPanel)
    if (xPlayer < MIN_X)
      xPlayer = MIN_X;
    else if (xPlayer > MAX_X)
      xPlayer = MAX_X;

    if (yPlayer < MIN_Y)
      yPlayer = MIN_Y;
    else if (yPlayer > MAX_Y)
      yPlayer = MAX_Y;

    repaint();
  }  // end of move()


  private void turn(int ang)
  // rotate the player by an ang degrees
  {
    angle = (angle + ang) % 360;
    repaint();
  } // end of turn()


  // ---------------- drawing the scene --------------------------


  public void paintComponent(Graphics g)
  // draw the land and then the player
  {
    super.paintComponent(g);

    landMap.draw(g, xPlayer, yPlayer, angle);     // draw rotated land image

    g.drawImage(playerIm, xDrawPos, yDrawPos, null);  // draw player
  }  // end of paintComponent()


}  // end of TourRotLandPanel class
